@@ -7,12 +7,14 @@ from traitlets import Unicode, Integer
 import traitlets
 import logging
 import numpy as np
-from .serialize import array_cube_png_serialization, array_serialization, array_sequence_serialization,\
+from .serialize import array_cube_tile_serialization, array_serialization, array_sequence_serialization,\
     color_serialization, image_serialization, texture_serialization
 from .transferfunction import *
+from .utils import debounced, grid_slice, reduce_size
 import warnings
 import ipyvolume
 import ipywebrtc
+import pythreejs
 
 logger = logging.getLogger("ipyvolume")
 
@@ -21,7 +23,7 @@ import ipyvolume._version
 semver_range_frontend = "~" + ipyvolume._version.__version_js__
 
 @widgets.register
-class Mesh(widgets.DOMWidget):
+class Mesh(widgets.Widget):
     _view_name = Unicode('MeshView').tag(sync=True)
     _view_module = Unicode('ipyvolume').tag(sync=True)
     _model_name = Unicode('MeshModel').tag(sync=True)
@@ -51,13 +53,21 @@ class Mesh(widgets.DOMWidget):
 #                                     default_value="green").tag(sync=True)
 #    geo = traitlets.Unicode('diamond').tag(sync=True)
     visible = traitlets.CBool(default_value=True).tag(sync=True)
-    visible_lines = traitlets.CBool(default_value=True).tag(sync=True)
-    visible_faces = traitlets.CBool(default_value=True).tag(sync=True)
 
-    side = traitlets.CaselessStrEnum(['front', 'back', 'both'], 'both').tag(sync=True)
+    material = traitlets.Instance(pythreejs.ShaderMaterial, help='A :any:`pythreejs.ShaderMaterial` that is used for the mesh')\
+                                  .tag(sync=True, **ipywidgets.widget_serialization)
+    @traitlets.default('material')
+    def _default_material(self):
+        return pythreejs.ShaderMaterial(side=pythreejs.Side.DoubleSide)
+
+    line_material = traitlets.Instance(pythreejs.ShaderMaterial, help='A :any:`pythreejs.ShaderMaterial` that is used for the lines/wireframe')\
+                                       .tag(sync=True, **ipywidgets.widget_serialization)
+    @traitlets.default('line_material')
+    def _default_line_material(self):
+        return pythreejs.ShaderMaterial()
 
 @widgets.register
-class Scatter(widgets.DOMWidget):
+class Scatter(widgets.Widget):
     _view_name = Unicode('ScatterView').tag(sync=True)
     _view_module = Unicode('ipyvolume').tag(sync=True)
     _model_name = Unicode('ScatterModel').tag(sync=True)
@@ -85,10 +95,92 @@ class Scatter(widgets.DOMWidget):
     geo = traitlets.Unicode('diamond').tag(sync=True)
     connected = traitlets.CBool(default_value=False).tag(sync=True)
     visible = traitlets.CBool(default_value=True).tag(sync=True)
-    visible_lines = traitlets.CBool(default_value=False).tag(sync=True)
-    visible_markers = traitlets.CBool(default_value=True).tag(sync=True)
 
+    texture = traitlets.Union([
+        traitlets.Instance(ipywebrtc.MediaStream),
+        Unicode(),
+        traitlets.List(Unicode, [], allow_none=True),
+        Image(default_value=None, allow_none=True),
+        traitlets.List(Image(default_value=None, allow_none=True))
+    ]).tag(sync=True, **texture_serialization)
 
+    material = traitlets.Instance(pythreejs.ShaderMaterial, help='A :any:`pythreejs.ShaderMaterial` that is used for the mesh')\
+                                  .tag(sync=True, **ipywidgets.widget_serialization)
+    @traitlets.default('material')
+    def _default_material(self):
+        return pythreejs.ShaderMaterial()
+
+    line_material = traitlets.Instance(pythreejs.ShaderMaterial, help='A :any:`pythreejs.ShaderMaterial` that is used for the lines/wireframe')\
+                                       .tag(sync=True, **ipywidgets.widget_serialization)
+    @traitlets.default('line_material')
+    def _default_line_material(self):
+        return pythreejs.ShaderMaterial()
+
+@widgets.register
+class Volume(widgets.Widget):
+    """Widget class representing a volume (rendering) using three.js"""
+    _view_name = Unicode('VolumeView').tag(sync=True)
+    _view_module = Unicode('ipyvolume').tag(sync=True)
+    _model_name = Unicode('VolumeModel').tag(sync=True)
+    _model_module = Unicode('ipyvolume').tag(sync=True)
+    _view_module_version = Unicode(semver_range_frontend).tag(sync=True)
+    _model_module_version = Unicode(semver_range_frontend).tag(sync=True)
+
+    data = Array(default_value=None, allow_none=True).tag(sync=True, **array_cube_tile_serialization)
+    data_original = Array(default_value=None, allow_none=True)
+    data_max_shape = traitlets.CInt(None, allow_none=True)  # TODO: allow this to be a list
+    data_min = traitlets.CFloat(0).tag(sync=True)
+    data_max = traitlets.CFloat(1).tag(sync=True)
+    show_min = traitlets.CFloat(0).tag(sync=True)
+    show_max = traitlets.CFloat(1).tag(sync=True)
+    clamp_min = traitlets.CBool(False).tag(sync=True)
+    clamp_max = traitlets.CBool(False).tag(sync=True)
+    opacity_scale = traitlets.CFloat(1.0).tag(sync=True)
+    brightness = traitlets.CFloat(1.0).tag(sync=True)
+    tf = traitlets.Instance(TransferFunction, allow_none=True).tag(sync=True, **ipywidgets.widget_serialization)
+    ray_steps = traitlets.CInt(None, allow_none=True, help='defines the length of the ray (1/ray_steps) for each step, in normalized coordintes.').tag(sync=True)
+
+    rendering_method = traitlets.Enum(values=['NORMAL', 'MAX_INTENSITY'], default_value='NORMAL').tag(sync=True)
+    lighting = traitlets.Bool(True).tag(sync=True)
+
+    extent = traitlets.Any().tag(sync=True)
+    extent_original = traitlets.Any()
+
+    def __init__(self, **kwargs):
+        super(Volume, self).__init__(**kwargs)
+        self._update_data()
+        self.observe(self.update_data, ['data_original', 'data_max_shape'])
+
+    def _listen_to(self, fig):
+        fig.observe(self.update_data, ['xlim', 'ylim', 'zlim'])
+
+    @debounced(method=True)
+    def update_data(self, change=None):
+        self._update_data()
+
+    def _update_data(self):
+        if self.data_original is None:
+            return
+        if all([k <= self.data_max_shape for k in self.data_original.shape]):
+            self.data = self.data_original
+            self.extent = self.extent_original
+            return
+        import ipyvolume as ipv
+        current_figure = ipv.gcf()
+        xlim = current_figure.xlim
+        ylim = current_figure.ylim
+        zlim = current_figure.zlim
+        shape = self.data_original.shape
+        ex = self.extent_original
+        viewx, xt = grid_slice(ex[0][0], ex[0][1], shape[2], *xlim)
+        viewy, yt = grid_slice(ex[1][0], ex[1][1], shape[1], *ylim)
+        viewz, zt = grid_slice(ex[2][0], ex[2][1], shape[0], *zlim)
+        view = [slice(*viewz), slice(*viewy), slice(*viewx)]
+        data_view = self.data_original[view]
+        extent = [xt, yt, zt]
+        data_view, extent = reduce_size(data_view, self.data_max_shape, extent)
+        self.data = np.array(data_view)
+        self.extent = extent
 
 @widgets.register
 class Figure(ipywebrtc.MediaStream):
@@ -100,26 +192,20 @@ class Figure(ipywebrtc.MediaStream):
     _view_module_version = Unicode(semver_range_frontend).tag(sync=True)
     _model_module_version = Unicode(semver_range_frontend).tag(sync=True)
 
-    volume_data = Array(default_value=None, allow_none=True).tag(sync=True, **array_cube_png_serialization)
     eye_separation = traitlets.CFloat(6.4).tag(sync=True)
-    data_min = traitlets.CFloat().tag(sync=True)
-    data_max = traitlets.CFloat().tag(sync=True)
-    tf = traitlets.Instance(TransferFunction, allow_none=True).tag(sync=True, **ipywidgets.widget_serialization)
-    anglex = traitlets.Float(0.0).tag(sync=True)
-    angley = traitlets.Float(0.0).tag(sync=True)
-    anglez = traitlets.Float(0.0).tag(sync=True)
-    angle_order = Unicode(default_value="XYZ").tag(sync=True)
 
     scatters = traitlets.List(traitlets.Instance(Scatter), [], allow_none=False).tag(sync=True, **ipywidgets.widget_serialization)
     meshes = traitlets.List(traitlets.Instance(Mesh), [], allow_none=False).tag(sync=True, **ipywidgets.widget_serialization)
+    volumes = traitlets.List(traitlets.Instance(Volume), [], allow_none=False).tag(sync=True, **ipywidgets.widget_serialization)
 
     animation = traitlets.Float(1000.0).tag(sync=True)
-    animation_exponent = traitlets.Float(.5).tag(sync=True)
+    animation_exponent = traitlets.Float(1.0).tag(sync=True)
 
     ambient_coefficient = traitlets.Float(0.5).tag(sync=True)
     diffuse_coefficient = traitlets.Float(0.8).tag(sync=True)
     specular_coefficient = traitlets.Float(0.5).tag(sync=True)
     specular_exponent = traitlets.Float(5).tag(sync=True)
+
     stereo = traitlets.Bool(False).tag(sync=True)
 
     camera_control = traitlets.Unicode(default_value='trackball').tag(sync=True)
@@ -127,9 +213,27 @@ class Figure(ipywebrtc.MediaStream):
     camera_center = traitlets.List(traitlets.CFloat, default_value=[0, 0, 0]).tag(sync=True)
     #Tuple(traitlets.CFloat(0), traitlets.CFloat(0), traitlets.CFloat(0)).tag(sync=True)
 
+    camera = traitlets.Instance(pythreejs.Camera, help='A :any:`pythreejs.Camera` instance to control the camera')\
+                                .tag(sync=True, **ipywidgets.widget_serialization,)
+    @traitlets.default('camera')
+    def _default_camera(self):
+        # return pythreejs.CombinedCamera(fov=46, position=(0, 0, 2), width=400, height=500)
+        return pythreejs.PerspectiveCamera(fov=46, position=(0, 0, 2), width=400, height=500)
+
+    scene = traitlets.Instance(pythreejs.Scene).tag(sync=True, **ipywidgets.widget_serialization)
+    @traitlets.default('scene')
+    def _default_scene(self):
+        # could be removed when https://github.com/jovyan/pythreejs/issues/176 is solved
+        # the default for pythreejs is white, which leads the volume rendering pass to make everything white
+        return pythreejs.Scene(background=None)
+
     width = traitlets.CInt(500).tag(sync=True)
     height = traitlets.CInt(400).tag(sync=True)
     downscale = traitlets.CInt(1).tag(sync=True)
+    displayscale = traitlets.CFloat(1).tag(sync=True)
+    capture_fps = traitlets.CFloat(None, allow_none=True).tag(sync=True)
+    cube_resolution = traitlets.CInt(512).tag(sync=True)
+
     show = traitlets.Unicode("Volume").tag(sync=True) # for debugging
 
     xlim = traitlets.List(traitlets.CFloat, default_value=[0, 1], minlen=2, maxlen=2).tag(sync=True)
@@ -146,6 +250,10 @@ class Figure(ipywebrtc.MediaStream):
     style = traitlets.Dict(default_value=ipyvolume.styles.default).tag(sync=True)
 
     render_continuous = traitlets.Bool(False).tag(sync=True)
+    selector = traitlets.Unicode(default_value='lasso').tag(sync=True)
+    selection_mode = traitlets.Unicode(default_value='replace').tag(sync=True)
+    mouse_mode = traitlets.Unicode(default_value='normal').tag(sync=True)
+    panorama_mode = traitlets.Enum(values=['no', '360', '180'], default_value='no').tag(sync=True)
 
     #xlim = traitlets.Tuple(traitlets.CFloat(0), traitlets.CFloat(1)).tag(sync=True)
     #y#lim = traitlets.Tuple(traitlets.CFloat(0), traitlets.CFloat(1)).tag(sync=True)
@@ -154,9 +262,30 @@ class Figure(ipywebrtc.MediaStream):
     def __init__(self, **kwargs):
         super(Figure, self).__init__(**kwargs)
         self._screenshot_handlers = widgets.CallbackDispatcher()
-        self._lasso_handlers = widgets.CallbackDispatcher()
+        self._selection_handlers = widgets.CallbackDispatcher()
         self.on_msg(self._handle_custom_msg)
 
+    def __enter__(self):
+        """Sets this figure as the current in the pylab API
+
+        Example:
+        >>> f1 = ipv.figure(1)
+        >>> f2 = ipv.figure(2)
+        >>> with f1:
+        >>>  ipv.scatter(x, y, z)
+        >>> assert ipv.gcf() is f2
+        """
+
+
+        import ipyvolume as ipv
+        self._previous_figure = ipv.gcf()
+        ipv.figure(self)
+
+    def __exit__(self, type, value, traceback):
+        import ipyvolume as ipv
+        ipv.figure(self._previous_figure)
+        del self._previous_figure
+
     def screenshot(self, width=None, height=None, mime_type='image/png'):
         self.send({'msg':'screenshot', 'width':width, 'height':height, 'mime_type':mime_type})
 
@@ -166,49 +295,21 @@ class Figure(ipywebrtc.MediaStream):
     def _handle_custom_msg(self, content, buffers):
         if content.get('event', '') == 'screenshot':
             self._screenshot_handlers(content['data'])
-        if content.get('event', '') == 'lasso':
-            self._lasso_handlers(content['data'])
+        elif content.get('event', '') == 'selection':
+            self._selection_handlers(content['data'])
 
-    def on_lasso(self, callback, remove=False):
-        self._lasso_handlers.register_callback(callback, remove=remove)
+    def on_selection(self, callback, remove=False):
+        self._selection_handlers.register_callback(callback, remove=remove)
 
     def project(self, x, y, z):
         W = np.matrix(self.matrix_world).reshape((4,4))     .T
         P = np.matrix(self.matrix_projection).reshape((4,4)).T
         M = np.dot(P, W)
-        vertices = np.array([x, y, z, np.ones(len(x))])#.T
-        p = np.array(np.dot(M, vertices))
-        p = p / p[3]
-        return p
-
-def _volume_widets(v, lighting=False):
-    import ipywidgets
-    #angle1 = ipywidgets.FloatSlider(min=0, max=np.pi*2, value=v.angle1, description="angle1")
-    #angle2 = ipywidgets.FloatSlider(min=0, max=np.pi*2, value=v.angle2, description="angle2")
-    #ipywidgets.jslink((v, 'angle1'), (angle1, 'value'))
-    #ipywidgets.jslink((v, 'angle2'), (angle2, 'value'))
-    if lighting:
-        ambient_coefficient = ipywidgets.FloatSlider(min=0, max=1, step=0.001, value=v.ambient_coefficient, description="ambient")
-        diffuse_coefficient = ipywidgets.FloatSlider(min=0, max=1, step=0.001, value=v.diffuse_coefficient, description="diffuse")
-        specular_coefficient = ipywidgets.FloatSlider(min=0, max=1, step=0.001, value=v.specular_coefficient, description="specular")
-        specular_exponent = ipywidgets.FloatSlider(min=0, max=10, step=0.001, value=v.specular_exponent, description="specular exp")
-        #angle2 = ipywidgets.FloatSlider(min=0, max=np.pi*2, value=v.angle2, description="angle2")
-        ipywidgets.jslink((v, 'ambient_coefficient'), (ambient_coefficient, 'value'))
-        ipywidgets.jslink((v, 'diffuse_coefficient'), (diffuse_coefficient, 'value'))
-        ipywidgets.jslink((v, 'specular_coefficient'), (specular_coefficient, 'value'))
-        ipywidgets.jslink((v, 'specular_exponent'), (specular_exponent, 'value'))
-        widgets_bottom = [ipywidgets.HBox([ambient_coefficient, diffuse_coefficient]),
-         ipywidgets.HBox([specular_coefficient, specular_exponent])]
-    else:
-        widgets_bottom = []
-        v.ambient_coefficient = 1
-        v.diffuse_coefficient = 0
-        v.specular_coefficient = 0
-
-    return ipywidgets.VBox(
-        [v.tf.control(), v,
-         ] + widgets_bottom# , ipywidgets.HBox([angle1, angle2])
-    )
+        x = np.asarray(x)
+        vertices = np.array([x, y, z, np.ones(x.shape)])
+        screen_h = np.tensordot(M, vertices, axes=(1, 0))
+        xy = screen_h[:2] / screen_h[3]
+        return xy
 
 def volshow(*args, **kwargs):
     """Deprecated: please use ipyvolume.quickvolshow or use the ipyvolume.pylab interface"""
@@ -216,24 +317,20 @@ def volshow(*args, **kwargs):
     return quickvolshow(*args, **kwargs)
 
 def quickquiver(x, y, z, u, v, w, **kwargs):
-    import ipyvolume.pylab as p3
-    p3.figure()
-    p3.quiver(x, y, z, u, v, w, **kwargs)
-    return p3.current.container
+    import ipyvolume as ipv
+    ipv.figure()
+    ipv.quiver(x, y, z, u, v, w, **kwargs)
+    return ipv.gcc()
 
 def quickscatter(x, y, z, **kwargs):
-    import ipyvolume.pylab as p3
-    p3.figure()
-    p3.scatter(x, y, z, **kwargs)
-    return p3.current.container
+    import ipyvolume as ipv
+    ipv.figure()
+    ipv.scatter(x, y, z, **kwargs)
+    return ipv.gcc()
 
 
-def quickvolshow(data, lighting=False, data_min=None, data_max=None, tf=None, stereo=False,
-            width=400, height=500,
-            ambient_coefficient=0.5, diffuse_coefficient=0.8,
-            specular_coefficient=0.5, specular_exponent=5,
-            downscale=1,
-            level=[0.1, 0.5, 0.9], opacity=[0.01, 0.05, 0.1], level_width=0.1, **kwargs):
+def quickvolshow(data, lighting=False, data_min=None, data_max=None,  max_shape=256,
+            level=[0.1, 0.5, 0.9], opacity=[0.01, 0.05, 0.1], level_width=0.1, extent=None, memorder='C', **kwargs):
     """
     Visualize a 3d array using volume rendering
 
@@ -241,15 +338,8 @@ def quickvolshow(data, lighting=False, data_min=None, data_max=None, tf=None, st
     :param lighting: boolean, to use lighting or not, if set to false, lighting parameters will be overriden
     :param data_min: minimum value to consider for data, if None, computed using np.nanmin
     :param data_max: maximum value to consider for data, if None, computed using np.nanmax
-    :param tf: transfer function (see ipyvolume.transfer_function, or use the argument below)
-    :param stereo: stereo view for virtual reality (cardboard and similar VR head mount)
-    :param width: width of rendering surface
-    :param height: height of rendering surface
-    :param ambient_coefficient: lighting parameter
-    :param diffuse_coefficient: lighting parameter
-    :param specular_coefficient: lighting parameter
-    :param specular_exponent: lighting parameter
-    :param downscale: downscale the rendering for better performance, for instance when set to 2, a 512x512 canvas will show a 256x256 rendering upscaled, but it will render twice as fast.
+    :parap int max_shape: maximum shape for the 3d cube, if larger, the data is reduced by skipping/slicing (data[::N]), set to None to disable.
+    :param extent: list of [[xmin, xmax], [ymin, ymax], [zmin, zmax]] values that define the bounds of the volume, otherwise the viewport is used
     :param level: level(s) for the where the opacity in the volume peaks, maximum sequence of length 3
     :param opacity: opacity(ies) for each level, scalar or sequence of max length 3
     :param level_width: width of the (gaussian) bumps where the opacity peaks, scalar or sequence of max length 3
@@ -257,53 +347,11 @@ def quickvolshow(data, lighting=False, data_min=None, data_max=None, tf=None, st
     :return:
 
     """
-    if tf is None: # TODO: should this just call the pylab interface?
-        #tf = TransferFunctionJsBumps(**kwargs)
-        tf_kwargs = {}
-        # level, opacity and widths can be scalars
-        try:
-            level[0]
-        except:
-            level = [level]
-        try:
-            opacity[0]
-        except:
-            opacity = [opacity] * 3
-        try:
-            level_width[0]
-        except:
-            level_width = [level_width] * 3
-        #clip off lists
-        min_length = min(len(level), len(level_width), len(opacity))
-        level = list(level[:min_length])
-        opacity = list(opacity[:min_length])
-        level_width = list(level_width[:min_length])
-        # append with zeros
-        while len(level) < 3:
-            level.append(0)
-        while len(opacity) < 3:
-            opacity.append(0)
-        while len(level_width) < 3:
-            level_width.append(0)
-        for i in range(1,4):
-            tf_kwargs["level"+str(i)] = level[i-1]
-            tf_kwargs["opacity"+str(i)] = opacity[i-1]
-            tf_kwargs["width"+str(i)] = level_width[i-1]
-        tf = TransferFunctionWidgetJs3(**tf_kwargs)
-    if data_min is None:
-        data_min = np.nanmin(data)
-    if data_max is None:
-        data_max = np.nanmax(data)
-    v = Figure(volume_data=data, data_min=data_min, data_max=data_max, stereo=stereo,
-                            width=width, height=height,
-                            ambient_coefficient=ambient_coefficient,
-                            diffuse_coefficient=diffuse_coefficient,
-                            specular_coefficient=specular_coefficient,
-                            specular_exponent=specular_exponent,
-                            tf=tf, **kwargs)
-
-    box = _volume_widets(v, lighting=lighting)
-    return box
+    import ipyvolume as ipv
+    ipv.figure()
+    ipv.volshow(data, lighting=lighting, data_min=data_min, data_max=data_max, max_shape=max_shape,
+        level=level, opacity=opacity, level_width=level_width, extent=extent,  memorder=memorder, **kwargs)
+    return ipv.gcc()
 
 def scatter(x, y, z, color=(1,0,0), s=0.01):
     global _last_figure;
@@ -313,3 +361,13 @@ def scatter(x, y, z, color=(1,0,0), s=0.01):
     fig.scatter = Scatter(x=x, y=y, z=z, color=color, size=s)
     fig.volume.scatter = fig.scatter
     return fig
+
+# add all help strings to the __doc__ for the api docstrings
+for name, cls in list(vars().items()):
+    try:
+        if issubclass(cls, traitlets.HasTraits):
+            for trait_name, trait in cls.class_traits().items():
+                if 'help' in trait.metadata:
+                    trait.__doc__ = trait.metadata['help']
+    except TypeError:
+        pass
